-----------------------------------------------------------
-- Test Drive Unlimited 2 TERRAIN mesh script
-- for 3D STUDIO MAX. Made with 3DSMAX 9, script should
-- work on max 8+ . Some materals and normals will not be
-- correct. Made by vagos21. e-mail: vagos21@yahoo.com
-- 
-- Run the script and select the island area (eg area 2-3)
-- at first run, you might also want to let it convert all
-- the textures too. Wait for a few secs, and you're done!
-----------------------------------------------------------

struct objesctstable(name, offset)
struct materialstable(name, offset)
struct GEOM(name, numprims, prims = #() )
struct PRIM(vboffset, iboffset, numverts, numinds, amesh)

dir_array = #() --all island directories

macroScript TDU2_Terra category:"myscripts"
(
	undo off

	--
	islandBasePath = getSavePath caption:"Select folder of the island area you wish to import (e.g. area-2-3)"
	
	--get all subfolders
	getFilesRecursive islandBasePath
	
	convertTex = (queryBox "Do you want to convert and use all island sub-textures?" title: "Texture Conversion" )

	--filter sector folders and load meshes only for them
	for modelpath in dir_array do
	(
		words = filterString modelpath "\\"
		lastname = words[words.count]
		if ((substring lastname 1 6) == "sector") then
		(
			print ("importing " + lastname)
			
			--convert all textures in folder to DDS
			if convertTex then(convertTextures modelpath) 
				
			loadMesh modelpath --load terrain mesh
			buildMaterials modelpath --build and apply terrain material
		)
	)
	
) --main script

fn buildMaterials filename =
(
	foldername = filterString filename "\\"
	lastname = foldername[foldername.count]
	
	--load textures
	adiffusemap = Bitmaptexture fileName: (filename+"/maps/"+lastname+"-c.dds")
	aglossmap = Bitmaptexture fileName: (filename+"/maps/"+lastname+"-s.dds")
	
	--build material
	amaterial = standard showInViewport:true Name: lastname
	amaterial.diffuseMap=(adiffusemap)
	amaterial.specularLevelMap =(aglossmap )
	amaterial.twoSided = off	
	amaterial.ambient = color 0 0 0
	amaterial.Diffuse = color 128 128 128
	amaterial.Specular = color 255 255 255
	
	--assign material to terrain sector
	obj = getnodebyname lastname
	obj.material = amaterial
	
	print "Terrain imported!"
	
) --loadMaterials

fn loadMesh filename =
(
	--open file
	foldername = filterString filename "\\"
	lastname = foldername[foldername.count]
	path = (filename+"\\"+lastname+".3dg")
	f = fopen path "rb"
	
	--read number of meshes	
	fseek f 44 #seek_set
	meshesnum = ((readlong f #unsigned)-16) / 16

	--read objects table
	objects = #(meshesnum)
	for i=1 to meshesnum do
	(
		aname = ""
		for w=1 to 8 do (aname += (bit.intasChar(readByte f #unsigned)))
		aoffset = readlong f #unsigned
		fseek f 4 #seek_cur
		objects[i] = objesctstable aname aoffset
	)

	geoms = #() --meshes array
	
	--read GEOMS
	for i=1 to meshesnum do
	(
		primarr = #()
		fseek f (objects[i].offset + 8) #seek_set
		numprims = ((readlong f #unsigned)-16)/80
		fseek f 4 #seek_cur
		for p=1 to numprims do 
		(
				fseek f 16 #seek_cur
				iboffset = readlong f #unsigned
				vboffset = readlong f #unsigned
				fseek f 20 #seek_cur
				numverts = readlong f #unsigned
				fseek f 8 #seek_cur
				numinds = readlong f #unsigned
				fseek f 20 #seek_cur --jump remaining bytes
				aprim = PRIM vboffset iboffset numverts numinds
				append primarr aprim
		)
		ageom = GEOM objects[i].name numprims primarr
--print ageom
		append geoms ageom
	)

	
	farr = #() --faces array
	varr = #() --vertex positions array
	UVWarr = #() --uv coords array
	
	--read primitives as meshes and attach them together in a GEOM
	for i=1 to meshesnum do
	(
		for j=1 to geoms[i].numprims do
		(
			
			fseek f geoms[i].prims[j].vboffset #seek_set
			--jump a few bytes
			fseek f 16 #seek_cur
			--read number of vertices
			numverts = (readlong f #unsigned)
			--jump a few bytes
			fseek f 28 #seek_cur

			--read first set as points height
			for y = 0 to 256 do
			(
				for x = 0 to 256 do
				(
				--read positions
					avert = [3*x as float,-3*y as float,0.]
					avert.z = (readfloat f)
					append varr avert
				)
			)
			
			--rest of file is normals...
			
			--we have to automatically build faces array
			for ind = 1 to 256*257 do
			(
				if ((mod ind 257)==0 ) then
				(
					--do nothing
				)else
				(
					face1 = [ ind+1 , ind , ind+257 ]
					face2 = [ ind+1 , ind+257 , ind+1+257 ]
					append farr face1
					append farr face2
				)
			)
			
			--and also automatically build UV coords
			for indJ = 0 to 256 do
			(
				for indI = 0 to 256 do
				(
					u=(indI as float)/257.0 -- + 1./257.
					v = -(indJ as float)/257.0 -- - 1./256.
					aUVW = [u, v, 0.0]
					append UVWarr aUVW
				)
			)
			
		)
		
		--calculate mesh position according to its index
		split = filterString lastname "-"
		posy = lastname[lastname.count] as float
		posx = lastname[lastname.count-2] as float
		if ((mod posy 2) == 0) then
		(
			location = [posx*768.0-384, -posy*768.0, 0.0]
		)else(
			location = [posx*768.0, -posy*768.0, 0.0]
		)
		
		--build mesh from vertex and faces arrays	
		terramesh = mesh vertices:varr faces:farr name:lastname position:location
		
		--apply uv mapping
		applyuvw terramesh UVWarr
		

		
	)

	
	--close 3DG file
	fclose f
	
	--print "Meshes loaded!"
	
) --loadMeshes

fn convertTextures filename =
(
	--DDS HEADER info from http://msdn.microsoft.com/archive/default.asp?url=/archive/en-us/directx9_c/directx/graphics/reference/ddsfilereference/ddsfileformat.asp
	--DDS HEADER values from http://wiki.answers.com/Q/How_do_you_write_a_header_for_a_128x128x128_dds_volume_texture_file_with_no_mipmaps_and_does_it_have_to_be_written_in_binary
	
	--DDS header stuff
	DDSD_CAPS =			0x00000001
	DDSD_HEIGHT =		0x00000002
	DDSD_WIDTH =		0x00000004
	DDSD_PITCH =		0x00000008
	DDSD_PIXELFORMAT =	0x00001000
	DDSD_MIPMAPCOUNT =	0x00020000
	DDSD_MIPMAPCOUNT =	0x00020000
	DDSD_LINEARSIZE =	0x00080000
	DDSD_DEPTH =		0x00800000
	
	DDPF_ALPHAPIXELS =	0x00000001
	DDPF_FOURCC =		0x00000004
	DDPF_RGB =			0x00000040
	
	DDSCAPS_COMPLEX= 	0x00000008
	DDSCAPS_TEXTURE =	0x00001000
	DDSCAPS_MIPMAP =	0x00400000
	
	DDSCAPS2_CUBEMAP =			0x00000200
	DDSCAPS2_CUBEMAP_POSITIVEX=	0x00000400
	DDSCAPS2_CUBEMAP_NEGATIVEX=	0x00000800
	DDSCAPS2_CUBEMAP_POSITIVEY=	0x00001000
	DDSCAPS2_CUBEMAP_NEGATIVEY=	0x00002000
	DDSCAPS2_CUBEMAP_POSITIVEZ=	0x00004000
	DDSCAPS2_CUBEMAP_NEGATIVEZ=	0x00008000
	DDSCAPS2_VOLUME =			0x00200000
	
	DXT1 = 0x31545844
	DXT2 = 0x32545844
	DXT3 = 0x33545844
	DXT4 = 0x34545844
	DXT5 = 0x35545844
	
	ddsd = #(18) --dds description
	ddpf = #(8) --dds pixel format
	ddsCaps = #(4) --dds caps
	
	--open each texture
	for texfile in getFiles (filename+"\maps\*.2db") do
	(
		--read texture header stuff
		fin = fopen texfile "rb"
		--read texture size
		fseek fin 28 #seek_set
		texsize = (readlong fin #unsigned)-64
		--read texture name
		texname = ""
		for w=1 to 8 do (texname += (bit.intasChar(readByte fin #unsigned)))
		--read texture width and height
		width = readshort fin
		height = readshort fin
		fseek fin 2 #seek_cur
		--read number of mipmaps
		mipmapcount = readbyte fin #unsigned
		fseek fin 1 #seek_cur
		--read type of texture (DXT, uncompressed, cubemap etc...
		textype = readlong fin #unsigned
		fseek fin 28 #seek_cur
		--load texture data into array
		texdata = #(texsize)
		for i = 1 to texsize do texdata[i]= (readbyte fin #unsigned)
		--close source file
		fclose fin
		
		--get texture name without the 2db
		texnoend = substring texfile 1 (texfile.count-4) 
		print ("Converting texture: "+texnoend+".dds, Size: "+width as string+"x"+height as string+" mipmaps: "+mipmapcount as string)
		
		
		--build dds header from data
		dwMagic = 0x20534444 --"DDS"
		ddsd[1] = 124
		ddsd[3] = height
		ddsd[4] = width
		ddsd[6] = 32
		ddsd[7] = mipmapcount
		for i = 8 to 18 do ddsd[i]=0
		ddpf[1] = 32
		ddpf[4] = 32
		ddpf[5] = 0x00ff0000
		ddpf[6] = 0x0000ff00
		ddpf[7] = 0x000000ff
		ddpf[8] = 0xff000000
		ddsCaps[1] = DDSCAPS_MIPMAP
		ddsCaps[2] = 0
		ddsCaps[3] = 0
		ddsCaps[4] = 0
		--case specific header stuff
		case textype of
		(
			132:( --working
				--print "Texture format is DXT1 " --DXT1
				ddsd[2] = bit.or (bit.or (bit.or(bit.or (bit.or DDSD_CAPS DDSD_HEIGHT) DDSD_WIDTH) DDSD_PIXELFORMAT) DDSD_MIPMAPCOUNT) DDSD_LINEARSIZE
				ddpf[3] = DXT1
				ddsd[5] = texsize
				ddpf[2] = DDPF_FOURCC
				)
			136:( --working
				--print "Texture format is DXT5" --DXT5
				ddsd[2] = bit.or (bit.or (bit.or(bit.or (bit.or DDSD_CAPS DDSD_HEIGHT) DDSD_WIDTH) DDSD_PIXELFORMAT) DDSD_MIPMAPCOUNT) DDSD_LINEARSIZE
				ddpf[3] = DXT5
				ddsd[5] = texsize
				ddpf[2] = DDPF_FOURCC
				)
			144:(
				--print "Texture format is Uncompressed" --uncompressed RGBA!
				ddsd[2] = bit.or(bit.or(bit.or(bit.or (bit.or DDSD_CAPS DDSD_HEIGHT) DDSD_WIDTH) DDSD_PIXELFORMAT) DDSD_MIPMAPCOUNT) DDSD_PITCH
				ddpf[3] = 0
				ddsd[5] = width*4
				ddpf[2] = DDPF_RGB
				)
		)
	
		--open output file
		fout = fopen (texnoend+".dds") "wb"
		--write header to file
		writelong fout dwMagic
		for i = 1 to 18 do (writeLong fout ddsd[i])
		for i = 1 to 8  do (writeLong fout ddpf[i])
		for i = 1 to 4  do (writeLong fout ddsCaps[i])
		writeLong fout 0
		--write texture data to file
		for i = 1 to texsize do (writeByte fout texdata[i] #unsigned)
		--flush output file
		fflush fout
		--close output file
		fclose fout
	)
	print "Texture Conversion Complete!"
) --convertTextures

fn applyuvw anode auvarray	= --applies UVW to node
(
	if (auvarray.count > 0) then (
		anode.numtverts = anode.numverts
		buildtvfaces anode
		for vi = 1 to anode.numverts do (
			settvert anode vi auvarray[vi]
		)				
		allfaces = #()
		for fi = 1 to anode.numfaces do (
			crtface = getface anode fi
			append allfaces crtface
		)
		for fi = 1 to anode.numfaces do (
			settvface anode fi allfaces[fi]
		)
	)
) -- apply UV

fn getFilesRecursive root =
(
	dir_array = GetDirectories (root+"/*")
	for d in dir_array do
	(
		join dir_array (GetDirectories (d+"*"))
	)
)

